import base64
import math
import random
import time

from PyQt5.QtCore import pyqtSlot, QSize, Qt
from PyQt5.QtGui import QFont
from PyQt5.QtGui import QPixmap, QImage, QMovie
from PyQt5.QtWidgets import QLabel, QSizePolicy

from piveilance.config import Config
from piveilance.util import ImageManip
from piveilance.util import parse_type


class Layout:

    def __init__(self,
                 id="unidentified",
                 type='flow',
                 cols=3,
                 rows=3,
                 maxAllowed=0,
                 cameras=None):
        self.id = parse_type(id, str)
        self.type = parse_type(type, str)
        self.cols = parse_type(cols, int)
        self.rows = parse_type(rows, int)
        self.maxAllowed = max(parse_type(maxAllowed, int), 0)
        self.cameras = {v['id']: v for v in parse_type(cameras, list)}
        self.geometry = WindowGeometry(self.rows, self.cols)

        for c in self.cameras.values():
            c['position'] = parse_type(c.get('position'), tuple)
            c['order'] = parse_type(c.get('order'), int)

    def calculate(self):
        pass

    def build(self, camMap):
        pass

    def updateGeometry(self, width, height, camLimit):
        self.maxAllowed = self.maxAllowed or camLimit
        self.geometry.width = width
        self.geometry.height = height
        self.calculate()


class FlowLayout(Layout):

    def __init__(self, *args, **kwargs):
        super(FlowLayout, self).__init__(*args, **kwargs)
        self.adjustNumberAllowed = True

    def calculate(self):

        """
        # Iterative computation to determine actual cols

        """
        # Start by initial guess that ncols = ncams
        g = self.geometry
        cols = rows = frameSize = self.maxAllowed
        while cols > 0:
            rows = math.ceil(self.maxAllowed / cols)
            frameSize = g.width / cols
            if ((g.height - frameSize * rows) < frameSize
                    and (rows * cols - self.maxAllowed) <= 1
                    or (rows > cols and cols > 2)):
                break
            cols -= 1
        cols = 1 if cols < 1 else cols

        g.rows = rows
        g.cols = cols
        g.frameSize = frameSize
        g.calculateAllProperties()

    def build(self, camMap):

        cams = camMap.values()

        for c in cams:
            if c.id in self.cameras:
                c.order = self.cameras[c.id].get('order')

        cams = sorted(cams, key=lambda x: x.order or len(cams) + 1)
        cams = cams[0: self.maxAllowed]

        for i, c in enumerate(cams):
            c.position = self.geometry.grid[i]
        return {c.id: c for c in cams}


class FixedLayout(Layout):
    adjustNumberAllowed = False

    def __init__(self, *args, **kwargs):
        super(FixedLayout, self).__init__(*args, **kwargs)
        self.adjustNumberAllowed = False

    def calculate(self):
        """
        # Rows and cols are predefined from the config

        """
        self.geometry.frameSize = self.geometry.width / self.geometry.cols
        self.geometry.calculateAllProperties()

    def build(self, camMap):

        positioned = {}
        freeCams = {c for c in camMap.values() if c.id not in self.cameras}
        camsByPosition = {c.get('position'): c for c in self.cameras.values()}

        for pos in self.geometry.grid:
            if pos in camsByPosition:
                camId = camsByPosition[pos]['id']
                cam = camMap.get(camId)
                if not cam:
                    cam = PlaceholderCamera(id=camId)
            else:
                if freeCams:
                    cam = freeCams.pop()
                    camId = cam.id
                else:
                    camId = "undefined-{}".format(pos)
                    cam = PlaceholderCamera(id=camId, name="Offline")
            cam.position = pos
            positioned[camId] = cam
        return positioned


class View():

    def __init__(self,
                 id,
                 fullscreen,
                 stretch,
                 fontRatio,
                 labels,
                 showCoords):
        self.id = parse_type(id, str)
        self.fullscreen = parse_type(fullscreen, bool)
        self.stretch = parse_type(stretch, bool)
        self.fontRatio = parse_type(fontRatio, float)
        self.labels = parse_type(labels, bool)
        self.showCoords = parse_type(showCoords, bool)

    def getCamGlobals(self):
        return {
            'stretch': self.stretch,
            'fontRatio': self.fontRatio,
            'showLabels': self.labels,
            'showCoords': self.showCoords,
        }


class Camera(QLabel):
    def __init__(self, **kwargs):
        super(Camera, self).__init__(None)
        self.setStaticOptions(Config(kwargs))

    def setStaticOptions(self, options):
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.setMinimumSize(QSize(50, 50))
        self.label = QLabel()
        self.pixmap = None
        self.id = options['id']
        self.movie = None

        self.cropRatio = options.get_float('cropRatio', 0)
        if self.cropRatio < 0 or self.cropRatio > 1:
            raise ValueError("Crop cannot be negative or inverse (>1)")

        self.name = options.get('name', self.id)
        self.position = options.get('position', None)
        self.order = options.get('order', None)
        self.size = options.get_int('size', 50)
        self.showLabels = options.get_bool('showLabels', False)
        self.showCoords = options.get_bool('showCoords', True)
        self.fontRatio = options.get_float('fontRatio', 0.025)
        self.direction = options.get_string('direction', 'right')
        self.stretch = options.get_bool('stretch', False)
        self.setOptions({})

    @pyqtSlot(dict, name="updateCamOptions")
    def setOptions(self, options):
        # validate ?
        vars(self).update(options)
        self.setScaledContents(self.stretch)
        self.setFrameSize()
        self.setLabel()

    def setFrameSize(self):
        if self.pixmap:
            self.setPixmap(self.pixmap.scaled(self.size, self.size))

    def setLabel(self):
        # fit text if too big
        text = self.name if self.showLabels else ""
        if self.position and self.showCoords:
            text = str(self.position) + " " + text
        self.label.setText(text)
        self.label.setFont(QFont("Arial", self.fontRatio * self.size))
        self.label.setAlignment(Qt.AlignLeft | Qt.AlignTop)
        self.label.setStyleSheet("color: #05FF00; font-weight: bold")

    @pyqtSlot(object, name="setimage")
    def setImage(self, camData=None):
        pass


class PiCamera(Camera):
    def __init__(self, **kwargs):
        super(PiCamera, self).__init__(**kwargs)

    @pyqtSlot(object, name="setimage")
    def setImage(self, camData=None):
        if self.id in camData:
            data = self.getImage(camData[self.id]['image'])
            img = QImage()
            img.loadFromData(data)
            if self.cropRatio != 0:
                crop = max((img.width() - img.height()) * self.cropRatio, 0)
                img = ImageManip.crop_direction(img, crop, self.direction)

            time.sleep(0.005 * random.randint(0, 1))
            self.pixmap = QPixmap().fromImage(img)
            self.setFrameSize()

    def getImage(self, data):
        byte = data.encode('utf-8')
        return base64.decodebytes(byte)


class PlaceholderCamera(Camera):
    image = "resources/image/offline.gif"

    def __init__(self, **kwargs):
        super(PlaceholderCamera, self).__init__(**kwargs)
        self.setImage()

    def setFrameSize(self):
        if self.movie:
            self.movie.setScaledSize(QSize(self.size, self.size))

    def setImage(self, camData=None):
        self.movie = QMovie(self.image)
        self.setMovie(self.movie)
        self.setFrameSize()
        self.movie.start()


class WindowGeometry():

    def __init__(self, rows=0, cols=0):
        self.rows = rows
        self.cols = cols
        self.height = 0
        self.width = 0
        self.frameSize = 0
        self.margins = (0, 0, 0, 0)
        self.grid = []

    def total(self):
        return self.cols * self.rows

    def correctFrameSize(self):
        remainder_height = self.height - self.frameSize * self.rows
        if remainder_height < 0:
            self.frameSize = self.frameSize - abs(remainder_height / self.rows)

    def setMargins(self):
        rheight = max(self.height - self.frameSize * self.rows, 0) / 2
        vheight = max(self.width - self.frameSize * self.cols, 0) / 2
        self.margins = (vheight, rheight, vheight, rheight)

    def calculateAllProperties(self):
        self.grid = [(i, j) for i in range(self.rows) for j in range(self.cols)]
        self.correctFrameSize()
        self.setMargins()



import logging
import logging.config
import os
import re
import sys
from collections import UserDict, Mapping
from copy import deepcopy

import yaml

from piveilance.util import parse_type, correct_path, decode, as_list, parse_exception


class Config(UserDict):

    def __init__(self, data, merge_env=False):
        super(Config, self).__init__(data)
        if merge_env:
            self.env_overrides()

    def get_sub_config(self, name):
        try:
            config = self if name else {}
            for k in as_list(name):
                config = config[k]
        except KeyError as e:
            raise ValueError("Subkey does not exist: " + str(e))
        return Config(config)

    def get_as(self, name, default=None, required=False, as_type=None):
        val = self.get(name, default)
        if required and not val:
            raise AssertionError("Value for " + name + " required but not given")
        return parse_type(val, as_type) if as_type else val

    def get_bool(self, name, default=False, required=False):
        val = self.get_as(name, default, required)
        return str(val).lower() in ['true', '1']

    def get_string(self, name, default=None, required=False):
        return self.get_as(name, default, required, str)

    def get_int(self, name, default=None, required=False):
        return self.get_as(name, default, required, int)

    def get_float(self, name, default=None, required=False):
        return self.get_as(name, default, required, float)

    def get_list(self, name, default=None, required=False):
        return self.get_as(name, default, required, list)

    def get_dict(self, name, default=None, required=False):
        return self.get_as(name, default, required, dict)

    def merge(self, update):
        self.merge_dict(self.data, update)

    def get_path(self, name, default=None, required=False):
        raw = self.get_as(name, default, required)
        return os.sep.join(re.split('[/\\\\]+', raw)) if raw else None

    def load_data_for_keys(self, parent, dir=''):
        for r, path in parent.items():
            if path.split('.')[-1].lower() in ['yml', 'yaml']:
                parent[r] = self.load_yaml(os.path.join(dir, path))
            else:
                parent[r] = os.path.join(dir, path)

    def env_overrides(self, key="cfg."):
        overrides = {}
        reduced = {decode(k, True): decode(v, True) for k, v in os.environ._data.items() if decode(k, True)}
        reduced = {k.lstrip(key): v for k, v in reduced.items() if k.startswith(key)}

        for k, v in reduced.items():
            entry = Config.make_dict(k.split("."), v)
            Config.merge_dict(overrides, entry)
        Config.merge_dict(self.data, overrides)

    @classmethod
    def from_yaml(cls, props_file, merge_env=False):
        return Config(cls.load_yaml(props_file), merge_env)

    @staticmethod
    def load_yaml(yaml_file):
        with open(yaml_file, 'r') as file:
            return yaml.safe_load(file)

    @staticmethod
    def make_dict(keylist, value):
        tree_dict = {}
        for i, key in enumerate(reversed(keylist)):
            val = value if i == 0 else tree_dict
            tree_dict = {
                key: val}
        return tree_dict

    @staticmethod
    def merge_dict(d1, d2, immutable=False):
        """
        # Combine dictionaries recursively
        # preserving the originals
        # assumes d1 and d2 dictionaries!!
        :param d1: original dictionary
        :param d2: update dictionary
        :return:
        """

        d1 = {} if d1 is None else d1
        d2 = {} if d2 is None else d2
        d1 = deepcopy(d1) if immutable else d1

        for k in d2:
            # if d1 and d2 have dict for k, then recurse
            # else assign the new value to d1[k]
            if (k in d1 and isinstance(d1[k], Mapping)
                    and isinstance(d2[k], Mapping)):
                Config.merge_dict(d1[k], d2[k])
            else:
                d1[k] = d2[k]
        return d1


class ConfigLoader:

    def __init__(self, path):
        self.config = Config.from_yaml(path)

    def loadGlobalConfig(self):

        globalConfig = Config({})
        globalConfig['configuration'] = self.config.get_dict('configuration')
        globalConfig['cameras'] = self.get_iter_as_dict('cameras', 'id', required=True)
        globalConfig['generators'] = self.get_iter_as_dict('generators', 'id', required=True)
        globalConfig['views'] = self.get_iter_as_dict('views', 'id', required=True)
        globalConfig['layouts'] = self.get_iter_as_dict('layouts', 'id', required=True)
        return globalConfig

    def get_iter_as_dict(self, name, keyname, default=None, required=False):

        value = self.config.get(name, default)
        if not value and required:
            raise ValueError("'{}': config cannot be empty.".format(name))

        try:
            return {i[keyname]: i for i in value}
        except KeyError:
            raise ValueError("Invalid key attribute for '{}': '{}'".format(name, keyname))


class FormattedDumper(yaml.Dumper):
    @classmethod
    def options(cls):
        return {
            'Dumper': FormattedDumper,
            'default_flow_style': False,
            'sort_keys': False
        }

    def increase_indent(self, flow=False, indentless=False):
        return super(FormattedDumper, self).increase_indent(flow, False)

    @staticmethod
    def str_presenter(dumper, data):
        if len(data.splitlines()) > 1:  # check for multiline string
            return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='|')
        return dumper.represent_scalar('tag:yaml.org,2002:str', data)


yaml.add_representer(str, FormattedDumper.str_presenter)


class Logging:

    @staticmethod
    def init_logger(logging_config, logs_root):
        with open(logging_config) as cfg:
            config = yaml.safe_load(cfg)

        log_dirs = {}
        for k, h in config['handlers'].items():
            filename = h.get('filename')
            if filename:
                filename = correct_path(filename)
                if not filename == os.path.abspath(filename):
                    filename = os.path.join(logs_root, filename)
                h['filename'] = filename
                log_dir = os.path.dirname(filename)
                if not os.path.exists(log_dir):
                    os.makedirs(log_dir)
                log_dirs[k] = log_dir
            logging.config.dictConfig(config)
        return log_dirs

    @staticmethod
    def get_logger(name='main'):
        current = logging.getLoggerClass()
        logging.setLoggerClass(PiLogger)
        custom = logging.getLogger(name)
        logging.setLoggerClass(current)
        return custom


class PiLogger(logging.Logger):

    def __init__(self, name):
        logging.Logger.__init__(self, name)
        self.logger = super(PiLogger, self)

    def exc_detail(self, e, extra_msg='', chain=None):
        errors = []
        if chain:
            errors.extend([parse_exception(e) for e in as_list(chain)])
        errors.append(parse_exception(e, extra_msg))
        for e in errors:
            self.error("-----------------------")
            self.log_list(e, logging.ERROR)
            self.error("-----------------------")

    def log_list(self, list, level=logging.INFO):
        list = as_list(list)
        for l in list:
            super().log(level, l)

    def _log(self, level, msg, args, exc_info=None, extra=None, stack_info=False):
        super()._log(level, msg, args, exc_info=None, extra=None, stack_info=False)
        sys.stdout.flush()
        for h in logging.getLogger().handlers:
            h.flush()



import json
import time

import requests
from PyQt5.QtCore import QThread, pyqtSignal

from piveilance.model import PiCamera
from piveilance.util import parse_type


class Generator(QThread):
    updateCameras = pyqtSignal(object)

    def __init__(self, id=None, updateInterval=None, cameraRepo=None, parent=None, **kwargs):
        super(Generator, self).__init__(parent)
        self.id = parse_type(id, str)
        self.cameraRepo = parse_type(cameraRepo, dict)
        self.updateInterval = parse_type(updateInterval, float)

    def update(self):
        pass

    def run(self):
        while True:
            self.update()
            time.sleep(self.updateInterval)

    def createCamera(self, camId):
        # Fills in defaults for none fields
        try:
            defaultOptions = self.cameraRepo['default']
        except KeyError:
            print("Missing default camera settings - using builtin")
            defaultOptions = self.defaultCamera()
        cameraConfig = self.cameraRepo.get(camId)
        if cameraConfig:
            for k in defaultOptions:
                if cameraConfig.get(k) is None:
                    cameraConfig[k] = defaultOptions[k]
        else:
            cameraConfig = defaultOptions
            cameraConfig['id'] = camId
        cam = self.camType(**cameraConfig)
        self.updateCameras.connect(cam.setImage)
        return cam

    def defaultCamera(self):
        pass


class PiCamGenerator(Generator):

    def __init__(self, **kwargs):
        super(PiCamGenerator, self).__init__(**kwargs)
        self.dataUrl = parse_type(kwargs.get('dataUrl'), str)
        self.listRefresh = parse_type(kwargs.get('listRefresh', 10), float)
        self.camType = PiCamera

    def defaultCamera(self):
        # fall back in case missing from config
        return {
            'id': 'default',
            'type': 'PiCam',
            'crop_ratio': 0.2,
            'direction': 'right'
        }

    def update(self):
        try:
            img = requests.get(self.dataUrl)
            data = json.loads(img.content)
            camData = {v['source']: v for v in data}
            self.updateCameras.emit(camData)
        except Exception as e:
            print(str(e))  # need to handle

